import sys

if sys.platform == 'win32':
    from .win32crypt import CryptUnprotectData

    def decrypt(cipher_blob: bytes) -> bytes:
        return CryptUnprotectData(cipher_blob)[1]

else:
    from .unix_crypt import aes_cbc_decrypt, get_chrome_safe_storage

    def decrypt(cipher_blob: bytes) -> bytes:
        if not cipher_blob:
            return cipher_blob
        elif cipher_blob.startswith(b'v10'):
            password = b'peanuts'
        elif cipher_blob.startswith(b'v11'):
            password = get_chrome_safe_storage()
        else:
            raise ValueError('Unknown cipher type: %r' % cipher_blob)
        return aes_cbc_decrypt(cipher_blob, password)



# DPAPI access library
# This file uses code originally created by Crusher Joe:
# http://article.gmane.org/gmane.comp.python.ctypes/420

from ctypes import POINTER, Structure, c_char, c_wchar_p
from ctypes import byref, create_string_buffer, memmove, windll
from ctypes.wintypes import DWORD

CRYPTPROTECT_UI_FORBIDDEN = 0x01


class DATA_BLOB(Structure):
    _fields_ = [
        ('cbData', DWORD),
        ('pbData', POINTER(c_char))
    ]


def CryptUnprotectData(
    cipher_text=b'', entropy=b'', reserved=None, prompt_struct=None,
    flags=CRYPTPROTECT_UI_FORBIDDEN
):
    blob_in, blob_entropy, blob_out = map(
        lambda x: DATA_BLOB(len(x), create_string_buffer(x)),
        [cipher_text, entropy, b'']
    )
    desc = c_wchar_p()

    if not windll.crypt32.CryptUnprotectData(
        byref(blob_in), byref(desc), byref(blob_entropy),
        reserved, prompt_struct, flags, byref(blob_out)
    ):
        raise RuntimeError('Failed to decrypt the cipher text with DPAPI')

    description = desc.value
    buffer_out = create_string_buffer(int(blob_out.cbData))
    memmove(buffer_out, blob_out.pbData, blob_out.cbData)
    map(windll.kernel32.LocalFree, [desc, blob_out.pbData])
    return description, buffer_out.value



import datetime
import http.cookiejar
import sqlite3
from typing import Dict, Iterable, Optional, Any
from .decrypt import decrypt
from .path import find_cookies_path


class ChromeCookieJar(http.cookiejar.CookieJar):

    def __init__(
        self,
        cookies_path=None,
        host_filter='%',
        policy=None
    ):
        '''Create CookieJar instance from the Chrome cookies database.

        Parameters
        ----------
        cookies_path: str, optional
            Path of the Chrom(ium) Cookies database file.
            If omitted, try to find the file in its default locations.

        host_filter: str, optional
            Filter cookies in the database by their host names.
            Only matched items would be added to the cookie jar.
            Support % and _ wildcards, as in the SQL "LIKE" clause.
            If omitted, include all cookies in the database.
        '''
        if cookies_path is None:
            cookies_path = find_cookies_path()

        super().__init__(policy)
        dummy = {key: None for key in (
            'port', 'port_specified', 'domain_specified', 'domain_initial_dot',
            'path_specified', 'discard', 'comment', 'comment_url', 'rest'
        )}

        with sqlite3.connect(cookies_path) as conn:
            conn.row_factory = dict_factory
            sql_fields = ', '.join(self.get_cookie_fields(conn))
            sql = 'select %s from cookies where host_key like ?' % sql_fields
            for row in conn.execute(sql, [host_filter]):
                row['expires'] = nt_timestamp_to_unix(row.pop('expires_utc'))
                if row.get('encrypted_value') and not row.get('value'):
                    row['value'] = decrypt(row.pop('encrypted_value')).decode()
                self.set_cookie(http.cookiejar.Cookie(
                    **row, **dummy, version=0,  # typing: ignore
                ))

    @staticmethod
    def get_cookie_fields(conn: sqlite3.Connection) -> Iterable[str]:
        'Adjust table fields to be compatible with different Chrome versions.'
        yield from [
            'host_key as domain', 'name', 'value', 'path', 'expires_utc'
        ]
        fields = set(get_field_names(conn, table='cookies'))

        # Chrome encrypts cookie values since version 33
        if 'encrypted_value' in fields:
            yield 'encrypted_value'

        # 'secure' was renamed to 'is_secure' since version 66
        if 'is_secure' in fields:
            yield 'is_secure as secure'
        else:
            yield 'secure'


def dict_factory(cursor: sqlite3.Cursor, row: tuple) -> Dict[str, Any]:
    'This row factory returns each row as a dict, with field names as the key.'
    return {
        col[0]: row[idx]
        for idx, col in enumerate(cursor.description)
    }


def get_field_names(conn: sqlite3.Connection, table: str) -> Iterable[str]:
    'Return all field names in the specified database table.'
    for field in conn.execute('pragma table_info("%s")' % table):
        yield field['name']


def nt_timestamp_to_unix(nt_timestamp: int) -> Optional[float]:
    '''Convert Windows NT FILETIME timestamp to Unix epoch.

    >>> nt_timestamp_to_unix(10275638401000000)
    91111881601.0
    '''
    if not nt_timestamp:
        return None
    t = datetime.datetime(1601, 1, 1)
    t += datetime.timedelta(microseconds=nt_timestamp)
    return t.timestamp()



import os
import sys


def find_cookies_path() -> str:
    'Find the cookies database in default locations.'
    paths = DEFAULT_LOCATIONS.get(sys.platform, DEFAULT_LOCATIONS['linux'])
    for path in paths:
        file_path = os.path.join(os.path.expandvars(path), 'Cookies')
        if os.path.isfile(file_path):
            return file_path
    raise FileNotFoundError(
        'Cookies not found in the default locations. '
        'Please specify the path to the Cookies database.'
    )


DEFAULT_LOCATIONS = {
    'win32': [
        '%LOCALAPPDATA%/Google/Chrome/User Data/Default',
    ],
    'darwin': [
        '$HOME/Library/Application Support/Google/Chrome/Default',
        '$HOME/Library/Application Support/Chromium/Default',
    ],
    'linux': [
        '$HOME/.config/google-chrome/Default',
        '$HOME/.config/chromium/Default',
    ],
}



from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2


def aes_cbc_decrypt(
    cipher_blob: bytes,
    password: bytes,
    salt=b'saltysalt',
    length=16,
    iv=16 * b' ',
    iterations=1,
) -> bytes:
    key = PBKDF2(password, salt, length, iterations)  # type: ignore
    cipher = AES.new(key, AES.MODE_CBC, IV=iv)

    # strip off the 'v10' prefix before AES decrypt
    clear_text = cipher.decrypt(cipher_blob[3:])

    # strip off the padding
    padding = clear_text[-1] if clear_text else 0
    return clear_text[:-padding]


def get_chrome_safe_storage() -> bytes:
    raise NotImplementedError
